import numpy as np
from typing import List, Tuple, Dict
import scipy
from .utils import inv, get_nearest_PSD, permute, pdet, \
        revert_permute, preallocate, get_init_mat, LL_correct
from . import Filter

__all__ = ['Smoother']


class Smoother(object):
    """
    Given a filtered object, Smoother returns smoothed state estimation.
    Given an BSTS:

    xi_{t+1} = F_t * xi_t + B_t * x_t + v_t     (v_t ~ N(0, Qt))
    y_t = H_t * xi_t + D_t * x_t + w_t     (w_t ~ N(0, Rt))

    and initial conditions:

    xi_1_0 = E(xi_1) 
    P_1_0 = Cov(xi_1)

    We want to solve:

    xi_t_{t-1} = E(xi_t|Info(T))
    P_t_{t-1} = Cov(xi_t|Info(T))

    where Info(t) is the information set at time t, and T = max(t). 
    Using forward filtering then backward smoothing, we are able to 
    characterize the distribution of the BSTS based on the full 
    information set up to T. Refer to doc/manual.pdf for details.
    """

    def __init__(self) -> None:
        """
        Initialize a Kalman Smoother. self.delta2 and self.chi2 
        are used for EM algorithms later.
        """
        self.xi_t_T = None
        self.P_t_T = None
        self.xi2_t_T = None
        self.Pcov_t_t1 = None
        self.delta2 = None
        self.chi2 = None
        self.r0_t = None
        self.r1_t = None
        self.N0_t = None
        self.N1_t = None
        self.N2_t = None
        self.is_smoothed = False
        

    def init_attr_smoother(self, kf: Filter) -> None:
        """
        Initialize attributes for Kalman smoother

        Parameters:
        ----------
        kf : a filter instance
        """
        # Include filtered results if kf is fitted and kf.for_smoother is True
        if not kf.is_filtered:
            raise TypeError('The Kalman filter object is not fitted yet')
        if not kf.for_smoother:
            raise TypeError('The Kalman filter object is not for smoothers')
        self.__dict__.update(kf.__dict__)

        # Initiate state variables
        self.xi_t_T = preallocate(self.T)
        self.P_t_T = preallocate(self.T)
        self.xi2_t_T = preallocate(self.T)
        self.Pcov_t_t1 = preallocate(self.T)
        self.delta2 = preallocate(self.T)
        self.chi2 = preallocate(self.T)

        # Initiate r and N
        self.r0_t = preallocate(self.T)
        self.r1_t = preallocate(self.t_q)
        self.N0_t = preallocate(self.T)
        self.N1_t = preallocate(self.t_q)
        self.N2_t = preallocate(self.t_q)
        
        self.r0_t[self.T-1] = np.zeros([self.xi_length, 1])
        self.N0_t[self.T-1] = np.zeros([self.xi_length, self.xi_length])
        
        if self.t_q > 0:
            self.r1_t[self.t_q-1] = np.zeros([self.xi_length, 1])
            self.N1_t[self.t_q-1] = np.zeros([self.xi_length, self.xi_length])
            self.N2_t[self.t_q-1] = np.zeros([self.xi_length, self.xi_length])


    def fit(self, kf: Filter) -> None:
        """
        Run backward smoothing. 

        Parameters: 
        ----------
        kf : a filter instance
        """
        self.init_attr_smoother(kf)

        # Start backward smoothing
        for t in reversed(range(self.T)):
            if t >= self.t_q:
                self._sequential_smooth(t)
            else:
                self._sequential_smooth_diffuse(t)

        self.is_smoothed = True
            

    def _sequential_smooth(self, t: int) -> None:
        """
        Update Kalman Smoother at time t. Refer to doc/manual.pdf
        for details on the notation of each variables.

        Parameters:
        ----------
        t : time index
        """
        H_t = self.Ht_tilde[t]
        d_t = self.d_t[t]
        Upsilon = self.Upsilon_star_t[t]
        L_t = self.L_star_t[t]
        n_t = self.n_t[t]

        # Backwards iteration on r and N
        r_t_1i = self.r0_t[t]
        N_t_1i = self.N0_t[t]
        
        for i in reversed(range(n_t)):
            H_i_T = H_t[i:i+1].T
            r_t_1i = (H_i_T).dot(d_t[i]) / Upsilon[i] + \
                    (L_t[i].T).dot(r_t_1i)
            N_t_1i = (H_i_T).dot(H_i_T.T) / Upsilon[i] + \
                    (L_t[i].T).dot(N_t_1i).dot(L_t[i])

        # Update smoothed xi, P and Pcov
        xi_t_1 = self.xi_t[t][0]
        P_t_1 = self.P_star_t[t][0]
        xi_t_T = xi_t_1 + P_t_1.dot(r_t_1i)
        P_t_T = get_nearest_PSD(P_t_1 - P_t_1.dot(N_t_1i).dot(P_t_1))
        self.xi_t_T[t] = xi_t_T
        self.P_t_T[t] = P_t_T

        # Update r and N for current period
        self.r0_t[t] = r_t_1i
        self.N0_t[t] = N_t_1i
        
        # Update r, N and P_1t_t_T from t to t-1
        if t > 0:
            Pcov = self.P_star_t[t-1][self.n_t[t-1]].dot(self.Ft[t-1].T).dot(
                self.I - N_t_1i.dot(P_t_1))
            self.Pcov_t_t1[t-1] = Pcov
            self.r0_t[t-1] = (self.Ft[t-1].T).dot(r_t_1i)
            self.N0_t[t-1] = (self.Ft[t-1].T).dot(N_t_1i).dot(self.Ft[t-1])


    def _sequential_smooth_diffuse(self, t: int) -> None:
        """
        Update diffuse Kalman Smoother at time t. Refer to doc/manual.pdf
        for details on the notation of each variables.

        Parameters:
        ----------
        t : time index
        """
        d_t = self.d_t[t]
        H_t = self.Ht_tilde[t]
        Upsilon_inf = self.Upsilon_inf_t[t]
        Upsilon_star = self.Upsilon_star_t[t]
        L0_t = self.L0_t[t]
        L1_t = self.L1_t[t]
        L_star_t = self.L_star_t[t]
        gt_0 = self.Upsilon_inf_gt_0_t[t]
        n_t = self.n_t[t]

        # Backwards iteration on r and N
        r0_t_1i = self.r0_t[t]
        r1_t_1i = self.r1_t[t]
        N0_t_1i = self.N0_t[t]
        N1_t_1i = self.N1_t[t]
        N2_t_1i = self.N2_t[t]
        
        for i in reversed(range(n_t)):
            H_i_T = H_t[i:i+1].T
            
            # If Upsilon_inf > 0
            if gt_0[i]:
                
                # Must update r1 first, bc it uses r0_t_1i
                r1_t_1i = (H_i_T).dot(d_t[i]) / Upsilon_inf[i] + \
                        (L1_t[i].T).dot(r0_t_1i) + (L0_t[i].T).dot(r1_t_1i)
                r0_t_1i = (L0_t[i].T).dot(r0_t_1i)

                # Order of updating: N2->N1->N0
                L1N1L0 = (L1_t[i].T).dot(N1_t_1i).dot(L0_t[i])
                L1N0L0 = (L1_t[i].T).dot(N0_t_1i).dot(L0_t[i])
                N2_t_1i = -(H_i_T).dot(H_i_T.T) * Upsilon_star[i] / \
                        np.power(Upsilon_inf[i], 2) + L1N1L0 + L1N1L0.T + \
                        (L0_t[i].T).dot(N2_t_1i).dot(L0_t[i]) + \
                        (L1_t[i].T).dot(N0_t_1i).dot(L1_t[i])
                N1_t_1i = (H_i_T).dot(H_i_T.T) / Upsilon_inf[i] + \
                        L1N0L0 + L1N0L0.T + (L0_t[i].T).dot(N1_t_1i).dot(L0_t[i])
                N0_t_1i = (L0_t[i].T).dot(N0_t_1i).dot(L0_t[i])

            # If Upsilon_inf == 0
            else:
                r0_t_1i = (H_i_T).dot(d_t[i]) / Upsilon_star[i] + \
                        (L_star_t[i].T).dot(r0_t_1i)
                N0_t_1i = (H_i_T).dot(H_i_T.T) / Upsilon_star[i] + \
                        (L_star_t[i].T).dot(N0_t_1i).dot(L_star_t[i])
                N1_t_1i = (L_star_t[i].T).dot(N1_t_1i).dot(L_star_t[i])

        # Update smoothed xi and P
        xi_t_1 = self.xi_t[t][0]
        P_inf_t1 = self.P_inf_t[t][0]
        P_star_t1 = self.P_star_t[t][0]
        xi_t_T = xi_t_1 + P_star_t1.dot(r0_t_1i) + \
                P_inf_t1.dot(r1_t_1i)
        P_inf_N1 = P_inf_t1.dot(N1_t_1i) 
        P_inf_N1_P_star = P_inf_N1.dot(P_star_t1)
        P_star_N0 = P_star_t1.dot(N0_t_1i)
        P_star_N0_P_star = P_star_N0.dot(P_star_t1)
        P_inf_N2 = P_inf_t1.dot(N2_t_1i)
        P_inf_N2_P_inf = P_inf_N2.dot(P_inf_t1)
        P_star_N1 = P_star_t1.dot(N1_t_1i)
        P_t_T = get_nearest_PSD(P_star_t1 - P_inf_N1_P_star - \
                P_inf_N1_P_star.T - P_star_N0_P_star - P_inf_N2_P_inf)
        self.xi_t_T[t] = xi_t_T
        self.P_t_T[t] = P_t_T

        # Update r and N for current period
        self.r0_t[t] = r0_t_1i
        self.r1_t[t] = r1_t_1i
        self.N0_t[t] = N0_t_1i
        self.N1_t[t] = N1_t_1i
        self.N2_t[t] = N2_t_1i
        
        if t > 0:
            # Update P_1t_t_T
            Pcov = self.P_star_t[t-1][self.n_t[t-1]].dot(self.Ft[t-1].T).dot(
                    self.I - P_inf_N1.T - P_star_N0.T) - \
                    self.P_inf_t[t-1][self.n_t[t-1]].dot(self.Ft[t-1].T).dot(
                    P_inf_N2.T + P_star_N1.T)
            self.Pcov_t_t1[t-1] = Pcov
                        
            # Update r and N from t to t-1
            self.r0_t[t-1] = (self.Ft[t-1].T).dot(r0_t_1i)
            self.r1_t[t-1] = (self.Ft[t-1].T).dot(r1_t_1i)
            self.N0_t[t-1] = (self.Ft[t-1].T).dot(N0_t_1i).dot(self.Ft[t-1])
            self.N1_t[t-1] = (self.Ft[t-1].T).dot(N1_t_1i).dot(self.Ft[t-1])
            self.N2_t[t-1] = (self.Ft[t-1].T).dot(N2_t_1i).dot(self.Ft[t-1])


    def _E_delta2(self, Mt: List[np.ndarray], t: int) -> np.ndarray:
        """
        Calculated expected value of delta2. See doc/manual.pdf for details.

        Parameters:
        ----------
        Mt : system matrix from ft(theta)  # Note: not theta_i
        t : time index

        Returns:
        ----------
        delta2 : expectation term for xi in G
        """
        # For initial state, use xi_1_0 and P_1_0 instead
        if t == 0:
            delta = self.xi_t_T[t] - Mt['xi_1_0']
            delta2 = delta.dot(delta.T) + self.P_t_T[t]

        # For other state, use formular derived in doc/manual.pdf Appendix E
        else:
            delta = self.xi_t_T[t] - Mt['Ft'][t-1].dot(self.xi_t_T[t-1]) - \
                    Mt['Bt'][t-1].dot(self.Xt[t-1])
            FPcov = Mt['Ft'][t-1].dot(self.Pcov_t_t1[t-1])  # Pcov_t_t1 starts at t=1
            delta2 = delta.dot(delta.T) + self.P_t_T[t] - FPcov - FPcov.T + \
                    Mt['Ft'][t-1].dot(self.P_t_T[t-1]).dot(Mt['Ft'][t-1].T)

        return delta2


    def _E_chi2(self, Mt: List[np.ndarray], t: int) -> np.ndarray:
        """
        Calculate expected value of chi2. See doc/manual.pdf for details.

        Parameters:
        ----------
        Mt : system matrix from ft(theta)  # Note: not theta_i
        t : time index

        Returns:
        ----------
        chi2 : expectation term for y in G
        """
        n_t = self.n_t[t]
        R_t = self.Rt[t]
        y_t = self.Yt[t]

        # Ht and Dt from Mt that is parameterized by theta
        H_t_M = permute(Mt['Ht'][t], self.partition_index[t], 
                axis='row')[0:n_t]
        D_t_M = permute(Mt['Dt'][t], self.partition_index[t], 
                axis='row')[0:n_t]
        chi = y_t[0:n_t] - H_t_M.dot(self.xi_t_T[t]) - \
                D_t_M.dot(self.Xt[t])
        chi2 = chi.dot(chi.T) + H_t_M.dot(
                self.P_t_T[t]).dot(H_t_M.T)
        return chi2


    def G(self, theta) -> float:
        """
        Calculate G

        Parameters:
        ----------
        theta : parameters to feed in self.ft

        Returns:
        G : objective value for EM algorithms
        """
        Mt = self.ft(theta, self.T, **self.ft_kwargs)
        G1 = 0
        G2 = 0
       
        # Raise Error if the smoother object is not run first
        if not self.is_smoothed:
            raise ValueError('Smoother is not fitted.')

        for t in range(self.T):
            if t == 0:
                _, A, Pi, P_star_1 = get_init_mat(Mt['P_1_0'])
                G1 += np.log(pdet(P_star_1)) + scipy.trace(inv(
                    P_star_1).dot(self._E_delta2(Mt, t))) 
            else:
                G1 += np.log(pdet(Mt['Qt'][t-1])) + scipy.trace(inv(
                    Mt['Qt'][t-1]).dot(self._E_delta2(Mt, t)))
            
            if self.n_t[t] > 0:    
                # Sort Rt index
                R_t = permute(Mt['Rt'][t], self.partition_index[t], 
                        axis='both')[0:self.n_t[t], 0:self.n_t[t]]
                G2 += np.log(pdet(R_t)) + scipy.trace(inv(
                        R_t).dot(self._E_chi2(Mt, t)))
        
        G = G1 + G2

        # Only use marginal correction if non-explosive diffuse
        if (not self.explosive) and self.t_q > 0:
            G -= np.log(pdet(LL_correct(Mt['Ht'], Mt['Ft'],
                    self.n_t, A, index=self.partition_index)))

        return -G.item()


    def get_smoothed_val(self, is_xi: bool=True, xi_col: List[int]=None) \
            -> List[np.ndarray]:
        """
        Generated smoothed xi. If state is diffuse,
        no covariance for Yt. Use xi_col to include 
        only the important xi. 

        Parameters:
        ----------
        is_xi : whether returns xi values
        xi_col : column index of xi to be included. 

        Returns:
        ----------
        xi_t_T : smoothed state means
        P_t_T : smoothed state covariances
        """
        # Raise error if not fitted yet
        if not self.is_smoothed:
            raise TypeError('The Kalman smoother object is not fitted yet')

        # If xi_col is not specified, use all columns
        if xi_col is None:
            xi_col = list(range(self.xi_length))

        Yt_smoothed = preallocate(self.T)
        Yt_smoothed_cov = preallocate(self.T)
        xi_t_T = preallocate(self.T)
        P_t_T = preallocate(self.T)

        for t in range(self.T):
            # Get smoothed mean and cov of y_t
            Yt_smoothed[t], Yt_smoothed_cov[t] = \
                    self.get_smoothed_y(t)

            # Get xi if needed
            if is_xi:
                xi_t_T[t] = self.xi_t_T[t][xi_col]
                P_t_T[t] = self.P_t_T[t][xi_col][:, xi_col]
            
        return Yt_smoothed, Yt_smoothed_cov, xi_t_T, P_t_T

    
    def get_smoothed_y(self, t: int) -> Tuple[np.ndarray]:
        """
        Get smoothed mean and variance for y_t

        Parameters:
        ----------
        t : time index

        Returns:
        ----------
        y_t_T : smoothed y_t
        y_cov_T : covariance matrix for smoothed y_t
        """
        # If no missing, use y_t and 0
        if self.n_t[t] == self.y_length:
            y_t_T = self.Yt[t]
            y_cov_T = np.zeros([self.y_length, self.y_length])
 
        # if all missing, simple calculation from xi_t_T and P_t_T
        elif self.n_t[t] == 0: 
            y_t_T = self.Ht[t].dot(self.xi_t_T[t]) + \
                    self.Dt[t].dot(self.Xt[t])
            y_cov_T = self.Ht[t].dot(self.P_t_T[t]).dot(
                    self.Ht[t].T) + self.Rt[t]

        # If partially missing, need to rearrange the measurements
        else: 
            n_t = self.n_t[t]
            y1_t_T = self.Yt[t][0:n_t].copy()
            y_cov_T_permute = np.zeros([self.y_length, self.y_length])

            R_22 = self.Rt[t][n_t:][:, n_t:]
            R_21 = self.Rt[t][n_t:][:, :n_t]
            R_11 = self.Rt[t][:n_t][:, :n_t]
            B = R_21.dot(inv(R_11))
            H1 = self.Ht[t][:n_t]
            H2 = self.Ht[t][n_t:]
            D1 = self.Dt[t][:n_t]
            D2 = self.Dt[t][n_t:]

            # Get mean
            epsilon = y1_t_T - H1.dot(self.xi_t_T[t]) - D1.dot(self.Xt[t])
            y2_t_T = H2.dot(self.xi_t_T[t]) + D2.dot(self.Xt[t]) + B.dot(epsilon)
            y_t_T_permute = np.concatenate((y1_t_T, y2_t_T), axis=0) 

            # Get covariance
            R22_1 = R_22 - R_21.dot(inv(R_11)).dot(R_21.T)
            diff_H2 = H2 - B.dot(H1)
            y2_cov_T = diff_H2.dot(self.P_t_T[t]).dot(diff_H2.T) + \
                    R22_1
            y_cov_T_permute[n_t:][:, n_t:] = y2_cov_T

            # Restore to the original index
            original_index = revert_permute(self.partition_index[t])
            y_t_T = permute(y_t_T_permute, original_index)
            y_cov_T = permute(y_cov_T_permute, original_index, 
                    axis='both')
        
        return y_t_T, y_cov_T


    def get_filtered_state(self, t: int) -> Dict:
        """
        Call Filter.get_filtered_state

        Parameters:
        ----------
        t : time index

        Returns:
        ----------
        state_val : state info at time t
        """
        state_val = Filter.get_filtered_state(self, t)
        return state_val




import numpy as np
import pandas as pd
from typing import List, Any, Callable, Dict, Tuple
from ..core import Filter, Smoother
from ..core.utils import df_to_list, list_to_df, \
        simulated_data, get_diag, ft, create_col
import warnings
warnings.simplefilter('default')

__all__ = ['BaseOpt', 'BaseConstantModel']


class BaseOpt(object):
    """
    BaseOpt is the core of model solvers. It directly interacts 
    with the Kalman filters/smoothers, and can be inherited by 
    both constant M models and more complicated non-constant Mt models.
    """
    
    def __init__(self) -> None:
        """
        Initialize base model. 
        """
        self.ft = None
        self.ft_kwargs = None
        self.solver_kwargs = None
        self.theta_opt = None
        self.fval_opt = None
        self.x_col = None
        self.y_col = None
        self.solver = None

        # Store fitted Kalman smoother 
        self.ks_fitted = None


    def set_f(self, Ft: Callable, reset: bool=True, 
            **ft_kwargs) -> None:
        """
        Mapping from theta to M. Ft must be the form: 
        f: Ft(theta, T) -> [M_1, M_2,...,M_T]. 

        Parameters:
        ----------
        Ft : theta -> Mt
        ft_kwargs : kwargs for ft
        reset : if true, reset fitted values
        """
        self.ft = Ft
        self.ft_kwargs = ft_kwargs
        if reset:
            self.theta_opt = None
            self.fval_opt = None
            self.ks_fitted = None


    def set_solver(self, solver: Any, **solver_kwargs) -> None:
        """
        Get solver object for the model. The solver must be 
        solver(theta, obj, **kwargs) where theta is the paramter,
        obj is the objective function (e.g. likelihood), and **kwargs
        are kwargs for the solver object. The solver should return 
        optimal theta and optimal LL evaluation

        Parameters:
        ----------
        solver : a solver object
        kwargs : kwargs for solver
        """
        self.solver = solver
        self.solver_kwargs = solver_kwargs


    def fit(self, df: pd.DataFrame, theta_init: np.ndarray,
            y_col: List[str], x_col: List[str]=None, 
            method: str='LLY', EM_threshold: float=1e-3, 
            num_EM_iter: int=np.inf, post_min_iter: int=100, 
            EM_stopping_rate: float=0.01, init_state: Dict=None) -> None:
        """
        Fit the model and returns optimal theta. 

        Parameters:
        ----------
        df : input data to be fitted
        theta_init : initial theta. self.Ft(theta_init) produce 
            Mt for first iteration of EM algorithm.
        y_col : list of columns in df that belong to Yt
        x_col : list of columns in df that belong to Xt. May be None
        method : EM or LLY
        EM_threshold : stopping criteria
        num_EM_iter : number of iterations for EM algorithms
        post_min_iter : after a new minimum is set, terminate iteration 
            after the number of runs determined by this argument
        EM_stopping_rate : weight placed on counter, higher rate means 
            update theta_opt slower. EM_stopping_rate = 0 means full 
            weight on theta_opt after each iteration
        init_state : user-specified initial state values
        """
        # Raise exception if method is not correctly specified
        if method not in ['EM','LLY']:
            raise ValueError('method must be "EM", or "LLY".')

        # Raise exception if EM_stopping_rate is negative
        if EM_stopping_rate < 0:
            raise ValueError('EM_stopping_rate must be non-negative')

        # Raise exception if x_col or y_col is not list
        if x_col is not None:
            if not isinstance(x_col, list):
                raise TypeError('x_col must be a list.')

        if not isinstance(y_col, list):
            raise TypeError('y_col must be a list.')

        if self.ft is None:
            raise ValueError('Need ft')

        if self.solver is None:
            raise ValueError('Need solver')
        
        # Raise exception if Ft no callable
        if not isinstance(self.ft, Callable):
            raise TypeError('ft must be a function')

        # Pre-process data inputs
        self.x_col = x_col
        self.y_col = y_col

        # If x_col is given, convert dataframe to lists
        Xt = df_to_list(df, x_col)
        Yt = df_to_list(df, y_col)

        # Run solver
        if method == 'LLY': 
            obj = lambda theta: self.get_LLY(theta, Yt, Xt)
            self.theta_opt, self.fval_opt = self.solver(
                    theta_init, obj, **self.solver_kwargs)

        # Note that fval_opt in EM and in LLY are not comparable
        elif method == 'EM':
            dist = np.inf
            theta_i = theta_init
            counter = 0
            max_G = -np.inf
            clock = 0

            while (dist > EM_threshold) and (counter < num_EM_iter):
                kf = Filter(self.ft, for_smoother=True, **self.ft_kwargs)
                kf.fit(theta_i, Yt, Xt)
                ks = Smoother()
                ks.fit(kf)
                obj = ks.G
                theta_opt, G_opt = self.solver(theta_i, obj, 
                        **self.solver_kwargs)
                dist = np.abs(max_G - G_opt)
                
                # Record minimal G and reset clock
                if max_G < G_opt:
                    max_G = G_opt
                    clock = 0

                # Giving increasingly more weight on theta_i
                alpha = 1 / (1 + counter * EM_stopping_rate)
                theta_i = alpha * theta_opt + (1 - alpha) * theta_i
                counter += 1
                clock += 1

                if clock > post_min_iter:
                    warnings.warn('Premature termination of EM')
                    break

            self.theta_opt = theta_i
            self.fval_opt = ks.G(theta_i)

        # Generate fitted smoother
        kf = Filter(self.ft, **self.ft_kwargs, for_smoother=True)
        kf.fit(self.theta_opt, Yt, Xt, init_state=init_state)
        ks = Smoother()
        ks.fit(kf)
        self.ks_fitted = ks


    def get_LLY(self, theta: np.ndarray, Yt: List[np.ndarray], 
            Xt: List[np.ndarray]=None) -> float:
        """
        Wrapper for calculating LLY. Used as the objective 
        function for optimizers.

        Parameters:
        ----------
        theta : paratmers
        Yt : list of measurements
        Xt : list of regressors. May be None

        Returns:
        ----------
        lly : log likelihood from Kalman filters
        """
        kf = Filter(self.ft, **self.ft_kwargs)
        kf.fit(theta, Yt, Xt)
        return kf.get_LL()


    def predict_t(self, df: pd.DataFrame, theta: np.ndarray=None, 
            is_xi: bool=True, xi_col: List[int]=None, 
            init_state: Dict=None, t_index: int=0) -> \
            Tuple[pd.DataFrame, Smoother]: 
        """
        Wrapper for BaseOpt.predict. It set the index to t_index. The 
        value of t_index must be between BaseOpt.ks_fitted.t_q and 
        BaseOpt.ks_fitted.T

        Parameters:
        ----------
        df : df to be predicted. Use np.nan for missing Yt
        theta : override theta_opt using user-supplied theta
        is_xi : whether output xi
        xi_col : index of xi to be included
        init_state : user-specified initial state values
        t_index : starting time index of df related to training set.
            t_index == -1 refers to the BaseOpt.ks_fitted.T

        Returns:
        ----------
        df_fitted : Contains filtered/smoothed y_t, xi_t, and P_t
        """
        if init_state is None:
            t = self.ks_fitted.T if t_index == -1 else t_index
            init_state = self.ks_fitted.get_filtered_state(t)
        
        df_fitted = self.predict(df, theta=theta, is_xi=is_xi, 
                xi_col=xi_col, init_state=init_state)
        
        return df_fitted


    def predict(self, df: pd.DataFrame, theta: np.ndarray=None, 
            is_xi: bool=True, xi_col: List[int]=None, 
            init_state: Dict=None, fmean_suffix: str='_filtered',
            fvar_suffix: str='_fvar', smean_suffix: str='_smoothed',
            svar_suffix: str='_svar') -> pd.DataFrame: 
        """
        Predict time series. df should contain both training and 
        test data. If Yt is not available for some or all test data,
        use np.nan as placeholders. Accept user-supplied theta as well

        Parameters:
        ----------
        df : df to be predicted. Use np.nan for missing Yt
        theta : override theta_opt using user-supplied theta
        is_xi : whether output xi
        xi_col : index of xi to be included
        init_state : user-specified initial state values
        fmean_suffix : suffix to filtered mean variable
        fvar_suffix : suffix to filtered var variable
        smean_suffix : suffix to smoothed mean variable
        svar_suffix : suffix to smoothed var variable

        Returns:
        ----------
        df_fitted : Contains filtered/smoothed y_t, xi_t, and P_t
        """
        # Generate system matrices for prediction
        Xt = df_to_list(df, self.x_col)
        Yt = df_to_list(df, self.y_col)
        
        # Generate filtered predictions
        kf = Filter(self.ft, for_smoother=True, **self.ft_kwargs)

        # Override theta_opt if theta is not None
        if theta is not None:
            kf.fit(theta, Yt, Xt, init_state=init_state)
        elif self.theta_opt is not None:
            kf.fit(self.theta_opt, Yt, Xt, init_state=init_state)
        else:
            raise ValueError('Model is not fitted')
        
        # Fit smoother
        ks = Smoother()
        ks.fit(kf)
        
        # Get filtered Yt
        y_col_filter = create_col(self.y_col, suffix=fmean_suffix)
        y_filter_var = create_col(self.y_col, suffix=fvar_suffix)
        Yt_filtered, Yt_P, xi_t, P_t = kf.get_filtered_val(
                is_xi=is_xi, xi_col=xi_col)
        Yt_P_diag = get_diag(Yt_P)
        df_Yt_filtered = list_to_df(Yt_filtered, y_col_filter)
        df_Yt_fvar = list_to_df(Yt_P_diag, y_filter_var)
            
        # Get smoothed Yt
        y_col_smoother = create_col(self.y_col, suffix=smean_suffix)
        y_smoother_var = create_col(self.y_col, suffix=svar_suffix)
        Yt_smoothed, Yt_S, xi_T, P_T = ks.get_smoothed_val(
                is_xi=is_xi, xi_col=xi_col)
        Yt_S_diag = get_diag(Yt_S)
        df_Yt_smoothed = list_to_df(Yt_smoothed, y_col_smoother)
        df_Yt_svar = list_to_df(Yt_S_diag, y_smoother_var)

        # Generate xi values if needed
        if is_xi:
            if xi_col is None:
                xi_col = list(range(kf.xi_length))

            xi_col_f = ['xi_{}_filtered'.format(i) for i in xi_col]
            P_col_f = ['P_{}_filtered'.format(i) for i in xi_col]
            xi_col_s = ['xi_{}_smoothed'.format(i) for i in xi_col]
            P_col_s = ['P_{}_smoothed'.format(i) for i in xi_col]

            df_xi_t = list_to_df(xi_t, xi_col_f)
            P_t_diag = get_diag(P_t)
            df_P_t = list_to_df(P_t_diag, P_col_f)

            P_T_diag = get_diag(P_T)
            df_xi_T = list_to_df(xi_T, xi_col_s)
            df_P_T = list_to_df(P_T_diag, P_col_s)
            df_fs = pd.concat([df_Yt_filtered, df_Yt_fvar,
                               df_Yt_smoothed, df_Yt_svar, df_xi_t,
                               df_P_t, df_xi_T, df_P_T], axis=1)

        else:
            df_fs = pd.concat([df_Yt_filtered, df_Yt_fvar, 
                    df_Yt_smoothed, df_Yt_svar], axis=1)

        df_fs.set_index(df.index, inplace=True)

        # Warning if new columns are in the original df.col
        if len(set(df_fs.columns).intersection(set(df.columns))) > 0:
            warnings.warn('Input df and predicted df contain ' + \
                    'overlapping column names')

        df_fitted = pd.concat([df, df_fs], axis=1)
        return df_fitted


    def simulated_data(self, input_theta: np.ndarray=None, 
            Xt: pd.DataFrame=None, T: int=None, init_state: Dict=None) \
            -> Tuple[pd.DataFrame, List[str], List[str]]:
        """
        Calls utils.simulated_data

        Parameters:
        ----------
        input_theta : parameters of ft, if None, use self.theta_opt
        Xt : optional input deterministic values
        T : length of the dataframe
        init_state : user-specified initial state values

        Returns:
        ----------
        df : output dataframe
        y_col : column names of y_t
        xi_col : column names of xi_t
        """
        if input_theta is not None:
            theta_ = input_theta
        elif self.theta_opt is not None:
            theta_ = self.theta_opt
        else:
            raise ValueError('Model needs theta')
        
        if self.ft is None:
            raise ValueError('Model needs ft')

        df, y_col, xi_col = simulated_data(self.ft, theta_, Xt=Xt, 
                T=T, init_state=init_state, **self.ft_kwargs)
        return df, y_col, xi_col


class BaseConstantModel(BaseOpt):
    """
    Any BSTS model with constant system matrices may inherit this class.
    The child class should provide get_f function. It inherits from BaseOpt
    and has a customized get_f
    """

    def set_f(self, f: Callable, reset: bool=True, **ft_kwargs) -> None:
        """
        Mapping from theta to M. Provided by children classes.
        Must be the form of get_f(theta). If defined, it should
        return system matrix M

        Parameters:
        ----------
        f : theta -> M
        ft_kwargs : arguments for ft
        reset : if true, reset fitted values
        """
        ft_ = lambda theta, T, **ft_kwargs: \
                ft(theta, f, T, **ft_kwargs)
        super().set_f(ft_, reset=reset, **ft_kwargs)





import numpy as np
from typing import List, Callable, Tuple, Dict
import scipy.linalg as linalg
from copy import deepcopy 
from .utils import mask_nan, LL_correct, M_wrap, \
        min_val, pdet, check_consistence, get_init_mat, \
        permute, partition_index, gen_Xt, preallocate


__all__ = ['Filter']


class Filter(object):
    """
    Given a BSTS model:

    xi_{t+1} = F_t * xi_t + B_t * x_t + v_t     (v_t ~ N(0, Qt))
    y_t = H_t * xi_t + D_t * x_t + w_t     (w_t ~ N(0, Rt))

    and initial conditions:

    xi_1_0 = E(xi_1) 
    P_1_0 = Cov(xi_1) (P_1_0 may contain diffuse parts)

    We want to solve:

    xi_t_{t-1} = E(xi_t|Info(t-1))
    P_t_{t-1} = Cov(xi_t|Info(t-1))

    where Info(t) is the information set at time t. With Gaussian
    assumptions on v_t and w_t, we are able to characterize the
    distribution of the BSTS model. Refer to doc/manual.pdf for details.
    """

    def __init__(self, ft: Callable, for_smoother: bool=False, **kwargs) -> None:
        """
        Initialize a Kalman Filter. Filter take system matrices 
        Mt returns characteristics of the BSTS model. Note that self.Qt
        correspond to Q_{t+1}.

        Parameters:
        ----------
        ft : function that generate Mt
        for_smoother : whether to store extra information for Kalman smoothers
        kwargs : kwargs for ft
        """
        self.ft = ft
        self.for_smoother = for_smoother
        self.is_filtered = False  # determine if the filter object has been fitted
        self.ft_kwargs = kwargs

        # Create placeholders for other class attributes
        self.theta = None
        self.Ft = None
        self.Bt = None
        self.Ht = None
        self.Dt = None
        self.Qt = None
        self.Rt = None
        self.xi_length = None
        self.y_length = None
        self.Yt = None
        self.Xt = None
        self.T = None
        self.I = None
        self.P_star = None
        self.n_t = None
        self.xi_T1 = None
        self.P_T1 = None
        self.explosive = False  # default to be non-explosive

        # Create output matrices
        self.xi_1_0 = None
        self.P_1_0 = None
        self.xi_t = None
        self.Upsilon_inf_t = None
        self.Upsilon_star_t = None
        self.d_t = None
        self.P_inf_t = None
        self.P_star_t = None
        self.Upsilon_inf_gt_0_t = None  # store whether Upsilon_{\inf} > 0
        self.Pi = None
        self.A = None
        self.q = None
        self.t_q = 0

        # Create output matrices for smoothers
        if self.for_smoother:
            self.partition_index = None
            self.L0_t = None
            self.L1_t = None
            self.L_star_t = None
            self.Ht_tilde = None


    def init_attr(self, theta: np.ndarray, Yt: List[np.ndarray], 
            Xt: List[np.ndarray]=None, init_state: Dict=None) -> None:
        """
        Initialize inputs to the Kalman filter. 
    
        Parameters:
        ----------
        theta : input parameters
        Yt : input Yt from self.fit()
        Xt : input Xt from self.fit()
        init_state : user-specified initial state 
        """
        # Initialize data inputs
        self.theta = theta
        self.Yt = deepcopy(Yt)
        self.T = len(self.Yt)

        # Generate Mt and Xt, and populate system matrices of the BSTS model
        M_ = self.ft(self.theta, 1, **self.ft_kwargs)
        self.Xt = gen_Xt(Xt=Xt, B=M_['Bt'][0], T=self.T)

        # Check consistence
        Mt = self.ft(self.theta, self.T, **self.ft_kwargs)
        check_consistence(Mt, self.Yt[0], self.Xt[0], 
                init_state=init_state)

        self.Ft = M_wrap(Mt['Ft'])
        self.Bt = M_wrap(Mt['Bt'])
        self.Ht = M_wrap(Mt['Ht'])
        self.Dt = M_wrap(Mt['Dt'])
        self.Qt = M_wrap(Mt['Qt'])
        self.Rt = M_wrap(Mt['Rt'])
        self.xi_length = self.Ft[0].shape[0]
        self.y_length = self.Ht[0].shape[0]
        self.I = np.eye(self.xi_length)

        # Create output matrices
        self.xi_1_0 = Mt['xi_1_0']
        self.P_1_0 = Mt['P_1_0']

        # Check whether BSTS has explosive states
        eig = linalg.eigvals(self.Ft[0])
        self.explosive = np.any(np.abs(eig) > 1 + min_val)

        # Collect initialization information
        self.q, self.A, self.Pi, self.P_star = \
                get_init_mat(self.P_1_0)

        # If init_state is provided, override
        if init_state is not None:
            self.q = init_state.get('q', self.q)
            self.A = init_state.get('P_inf_t', self.A)
            self.P_star = init_state.get('P_star_t')
            self.xi_1_0 = init_state.get('xi_t', self.xi_1_0)
            self.explosive = init_state.get('explosive', 
                    self.explosive)

        # Initialize xi_1_0 and  P_1_0
        self.xi_t = preallocate(self.T, self.y_length + 1)
        self.xi_t[0][0] = self.xi_1_0
        self.d_t = preallocate(self.T, self.y_length)
        self.Upsilon_star_t = preallocate(self.T, self.y_length)
        self.P_star_t = preallocate(self.T, self.y_length + 1)
        self.P_star_t[0][0] = self.P_star
        self.n_t = preallocate(self.T)
        
        if self.q > 0:
            self.P_inf_t = preallocate(self.T, self.y_length + 1)
            self.P_inf_t[0][0] = self.A
            self.Upsilon_inf_t = preallocate(self.T, self.y_length)
            self.Upsilon_inf_gt_0_t = preallocate(self.T, self.y_length)

        if self.for_smoother:
            self.Ht_tilde = preallocate(self.T)
            self.partition_index = preallocate(self.T)
            self.L_star_t = preallocate(self.T, self.y_length)

            if self.q > 0:
                self.L0_t = preallocate(self.T, self.y_length)
                self.L1_t = preallocate(self.T, self.y_length)


    def fit(self, theta: np.ndarray, Yt: List[np.ndarray], 
            Xt: List[np.ndarray]=None, init_state: Dict=None) -> None:
        """
        Run forward filtering, given input measurements and regressors

        Parameters:
        ----------
        theta : list of parameters for self.ft
        Yt : measurements, may contain missing values
        Xt : regressors, must be deterministic and has no missing values.
            If set as None, will generate zero vectors
        init_state : user-specified initial state values
        """
        # Initialize input data
        self.init_attr(theta, Yt, Xt, init_state=init_state)

        # Filter
        for t in range(self.T):
            if self.q > 0:
                self._sequential_update_diffuse(t)
            
            else:
                self._sequential_update(t)

        # Mark the object as fitted
        self.is_filtered = True


    def _sequential_update(self, t: int) -> None:
        """
        Sequentially update Kalman Filter at time t. If Q_t
        is not diagonal, we first transform it to a diagonal
        matrix using LDL transformation. 

        Parameters:
        ----------
        t : time index
        """
        # LDL 
        n_t, Y_t, H_t, D_t, R_t, l_t, l_inv, partitioned_index = self._LDL(t)
        self.n_t[t] = n_t

        # Skip missing measurements
        for i in range(1, n_t+1):
            ob_index = i - 1  # state variables have one more value
            xi_i = self.xi_t[t][ob_index]
            P_i = self.P_star_t[t][ob_index]
            H_i = H_t[ob_index:i]
            D_i = D_t[ob_index:i]
            sigma2 = R_t[ob_index][ob_index]
            Upsilon = H_i.dot(P_i).dot(H_i.T) + sigma2
            d_t_i = Y_t[ob_index] - H_i.dot(xi_i) - D_i.dot(self.Xt[t])
            K = P_i.dot(H_i.T) / Upsilon
            xi_t_i1 = xi_i + K.dot(d_t_i)
            L_t_i = self.I - K.dot(H_i)
            KRK = K.dot(sigma2).dot(K.T)
            P_i1 = self._joseph_form(L_t_i, P_i, KRK)

            self.xi_t[t][i] = xi_t_i1
            self.P_star_t[t][i] = P_i1
            self.Upsilon_star_t[t][ob_index] = Upsilon
            self.d_t[t][ob_index] = d_t_i

            if self.for_smoother:
                self.L_star_t[t][ob_index] = L_t_i

        # Calculate xi_t1_t and P_t, and add placeholders for others
        xi_t1_1 = self.Ft[t].dot(self.xi_t[t][n_t]) + \
                self.Bt[t].dot(self.Xt[t])
        P_t1_1 = self.Ft[t].dot(self.P_star_t[t][n_t]).dot(
                self.Ft[t].T) + self.Qt[t]

        if t < self.T - 1:
            self.xi_t[t+1][0] = xi_t1_1
            self.P_star_t[t+1][0] = P_t1_1
        else:
            self.xi_T1 = xi_t1_1
            self.P_T1 = P_t1_1

        if self.for_smoother:
            self.Ht_tilde[t] = H_t
            self.partition_index[t] = partitioned_index


    def _sequential_update_diffuse(self, t: int) -> None:
        """
        Sequentially update diffuse kalman filters at time t. If Q_t 
        is not diagonal, we first transform it to a diagonal
        matrix using LDL transformation. Also update q_t. 
        All outputs are lists of numpy arrays. 
        
        Parameters:
        ----------
        t : time index
        """
        # LDL 
        n_t, Y_t, H_t, D_t, R_t, l_t, l_inv, partitioned_index = self._LDL(t)
        self.t_q += 1
        self.n_t[t] = n_t

        # Start sequential updating xi_{t:(i)}, 
        # P_inf_t_{t:(i)}, and P_star_t_{t:(i)}
        for i in range(1, n_t+1):
            ob_index = i - 1  # y index starts from 0 not 1
            P_inf = self.P_inf_t[t][ob_index]  # the most recent P
            P_star = self.P_star_t[t][ob_index]
            xi_i = self.xi_t[t][ob_index]
            H_i = H_t[ob_index:i]
            abs_Hi = np.abs(H_i)
            D_i = D_t[ob_index:i]
            sigma2 = R_t[ob_index][ob_index]
            Upsilon_inf = H_i.dot(P_inf).dot(H_i.T)
            Upsilon_star = H_i.dot(P_star).dot(H_i.T) + sigma2
            d_t_i = Y_t[ob_index] - H_i.dot(xi_i) - D_i.dot(self.Xt[t])    

            # If Upsilon_inf > 0
            gt_0 = Upsilon_inf > min_val * np.power(
                    abs_Hi[abs_Hi > min_val].min(), 2)
            if gt_0:
                K_0 = P_inf.dot(H_i.T) / Upsilon_inf
                K_1 = (P_star.dot(H_i.T) - K_0.dot(Upsilon_star)) / Upsilon_inf
                xi_t_i1 = xi_i + K_0.dot(d_t_i)
                L0_t_i = self.I - K_0.dot(H_i)
                L1_t_i = - K_1.dot(H_i)
                KRK = K_0.dot(sigma2).dot(K_0.T)
                P_inf_i1 = self._joseph_form(L0_t_i, P_inf)
                P_star_i1 = self._joseph_form(L0_t_i, P_star, KRK)

                if self.for_smoother:
                    self.L0_t[t][ob_index] = L0_t_i
                    self.L1_t[t][ob_index] = L1_t_i

                # update number of diffuse state
                self.q = self.q - 1

            # If Upsilon_inf == 0
            else:
                K_star = P_star.dot(H_i.T) / Upsilon_star
                xi_t_i1 = xi_i + K_star.dot(d_t_i)
                L_star_t_i = self.I - K_star.dot(H_i)
                KRK = K_star.dot(sigma2).dot(K_star.T)
                P_inf_i1 = P_inf
                P_star_i1 = self._joseph_form(L_star_t_i, P_star, KRK)
                
                if self.for_smoother:
                    self.L_star_t[t][ob_index] = L_star_t_i 

            self.xi_t[t][i] = xi_t_i1
            self.P_inf_t[t][i] = P_inf_i1
            self.P_star_t[t][i] = P_star_i1
            self.Upsilon_inf_t[t][ob_index] = Upsilon_inf
            self.Upsilon_star_t[t][ob_index] = Upsilon_star
            self.d_t[t][ob_index] = d_t_i
            self.Upsilon_inf_gt_0_t[t][ob_index] = gt_0

        # Calculate xi_t1_t, P_inf_t1_t, and P_star_t1_t, 
        # and add placeholders for others
        xi_t1_1 = self.Ft[t].dot(self.xi_t[t][n_t]) + \
                self.Bt[t].dot(self.Xt[t])
        P_inf_t1_1 = self.Ft[t].dot(
                self.P_inf_t[t][n_t]).dot(self.Ft[t].T)
        P_star_t1_1 = self.Ft[t].dot(
                self.P_star_t[t][n_t]).dot(self.Ft[t].T) + self.Qt[t]      

        # Update q at the end of time t
        self.q = min(self.q, np.linalg.matrix_rank(P_inf_t1_1))

        # Raise exception if we don't have enough data
        if t == self.T - 1:
            raise ValueError('Not enough data to handle diffuse priors')
        else:  # if no error raised, we are able to update the filter for t+1
            self.xi_t[t+1][0] = xi_t1_1
            self.P_inf_t[t+1][0] = P_inf_t1_1
            self.P_star_t[t+1][0] = P_star_t1_1 

        if self.for_smoother:
            self.Ht_tilde[t] = H_t
            self.partition_index[t] = partitioned_index


    def _joseph_form(self, L: np.ndarray, P: np.ndarray, 
            KRK: np.ndarray=None) -> np.ndarray:
        """
        Update P using Joseph Form, which guarantees 
        PSD of the covariance matrix. 

        Parameters: 
        ----------
        L : (I - K * d) 
        H : Hi
        P : state Covariance, can be either regular or diffuse
        KRK : for P_star or regular P, KRK is part of updating

        Returns:
        ----------
        P_1 : filtered P
        """
        if KRK is None:
            P_1 = L.dot(P).dot(L.T)
        else:
            P_1 = L.dot(P).dot(L.T) + KRK
        return P_1


    def _LDL(self, t: int) -> Tuple[int, np.ndarray, np.ndarray, \
            np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]: 
        """
        Transform the BSTS model using LDL methods. 

        Parameters:
        ----------
        t : time index of the BSTS system

        Returns:
        ----------
        n_t : number of observed measurement at time t
        Y_t : transformed Yt[t], NaN values are replaced with 0
        H_t : transformed Yt[t], rows that correspond to NaN 
            in Yt[t] are replaced with 0
        D_t : transformed Dt[t], rows that correspond to NaN 
            in Yt[t] are replaced with 0
        R_t : transformed Rt[t], rows and columns that correspond
            to NaN in Yt[t] are replaced with 0, then diagonalized
        L_t : l from ldl
        L_inv : l^{-1}, used for EM algorithms
        partitioned_index : sorted index, used for EM algorithms
        """
        # Preprocess Rt and Yt if Yt has missing measurements
        is_missing = np.hstack(np.isnan(self.Yt[t]))
        n_t = (~is_missing).sum()

        # Sort observed measurements to the top
        partitioned_index = partition_index(is_missing)  

        if np.any(is_missing):
            self.Yt[t] = mask_nan(is_missing, self.Yt[t], dim='row')
            self.Yt[t] = permute(self.Yt[t], partitioned_index)
            self.Rt[t] = permute(self.Rt[t], partitioned_index, axis='both')
            self.Ht[t] = permute(self.Ht[t], partitioned_index)
            self.Dt[t] = permute(self.Dt[t], partitioned_index)

        # Diagonalize Y_t, H_t, and D_t
        L_t, R_t, L_inv = self.Rt.ldl(t)
        Y_t = L_inv.dot(self.Yt[t])
        H_t = L_inv.dot(self.Ht[t])
        D_t = L_inv.dot(self.Dt[t])
        
        return n_t, Y_t, H_t, D_t, R_t, L_t, L_inv, partitioned_index


    def get_LL(self) -> float:
        """
        Calculate the marginal likelihood of Yt for a BSTS model.

        Returns:
        ----------
        LL : marginal likelihood 
        """
        LL = 0
        for t in range(self.T):
            Upsilon_star = self.Upsilon_star_t[t]
            d_t = self.d_t[t]

            # If diffuse, use Psi_t
            if t < self.t_q:
                gt_0 = self.Upsilon_inf_gt_0_t[t]
                Upsilon_inf = self.Upsilon_inf_t[t]
                for i in range(self.n_t[t]):
                    if gt_0[i]:
                        LL += np.log(Upsilon_inf[i])
                    else:
                        LL += np.log(Upsilon_star[i]) + (
                                d_t[i].T).dot(d_t[i]) / Upsilon_star[i]

            # If not diffuse, skip Psi_t
            else:
                for i in range(self.n_t[t]):
                    LL += np.log(Upsilon_star[i]) + (
                            d_t[i].T).dot(d_t[i]) / Upsilon_star[i]

        # Add marginal correction term if not explosive
        if (not self.explosive) and self.t_q > 0:
            LL -= np.log(pdet(LL_correct(self.Ht, self.Ft, 
                self.n_t, self.A)))
        
        return -LL.item()


    def get_filtered_val(self, is_xi: bool=True, xi_col: List[int]=None) \
            -> List[np.ndarray]:
        """
        Generated filtered Yt. It will also generate
        filtered values for missing measurements. If
        state is diffusal, no covariance for Yt. Use 
        xi_col to include only the important xi. For 
        example, we don't need all xi in models that
        have weekly effects.

        Parameters:
        ----------
        is_xi : whether returns xi values
        xi_col : column index of xi to be included. 

        Returns:
        ----------
        Yt_filtered : filtered Yt
        Yt_filtered_cov : covariance of filtered Yt
        xi_t : selected xi_t
        P_t : selected P_t
        """
        # Raise error if not fitted yet
        if not self.is_filtered:
            raise TypeError('The Kalman filter object is not fitted yet')

        # If xi_col is not specified, use all columns
        if xi_col is None:
            xi_col = list(range(self.xi_length))
        
        Mt = self.ft(self.theta, self.T, **self.ft_kwargs)

        Yt_filtered = preallocate(self.T)
        Yt_filtered_cov = preallocate(self.T)
        xi_t = preallocate(self.T)
        P_t = preallocate(self.T)

        for t in range(self.T):
            # Get filtered y_t
            Yt_filtered[t] = Mt['Ht'][t].dot(self.xi_t[t][0]) + \
                    Mt['Dt'][t].dot(self.Xt[t])

            # Get variance of filtered y_t
            if t >= self.t_q:
                Yt_filtered_cov[t] = Mt['Ht'][t].dot(self.P_star_t[t][0]).dot(
                        Mt['Ht'][t].T) + Mt['Rt'][t]
            else:
                Yt_filtered_cov[t] = np.nan * np.ones([self.y_length, self.y_length])

            # Get xi if needed
            if is_xi:
                xi_t[t] = self.xi_t[t][0][xi_col]
                if t >= self.t_q:
                    P_t[t] = self.P_star_t[t][0][xi_col][:, xi_col]
                else:
                    P_t[t] = np.nan * np.ones(self.P_star_t[t][0].shape)
            
        return Yt_filtered, Yt_filtered_cov, xi_t, P_t

    
    def get_filtered_state(self, t: int) -> Dict:
        """
        Get xi_t[t][0] and P_star_t[t][0]

        Parameters:
        ----------
        t : time index

        Returns:
        ----------
        state_val : state info at time t
        """
        if not self.is_filtered:
            raise ValueError('Kalman filter is not fitted yet')

        if (t >= self.t_q) and (t < self.T):
            xi_t = deepcopy(self.xi_t[t][0])
            P_star_t = deepcopy(self.P_star_t[t][0])
            P_inf_t = np.zeros(P_star_t.shape)
            q = 0
        
        # Successful training always ends with regular Kalman filters
        elif t == self.T:
            xi_t = deepcopy(self.xi_T1)
            P_star_t = deepcopy(self.P_T1)
            P_inf_t = np.zeros(P_star_t.shape)
            q = 0
        elif t > self.T:
            raise ValueError('Maximum t allowed is {}'.format(self.T))
        else:
            raise ValueError('Diffuse state at ' + \
                    'time {}'.format(t))
        
        state_val = {'xi_t': xi_t, 'P_star_t': P_star_t, 
                'P_inf_t': P_inf_t, 'q': q}
        return state_val



from .kalman_filter import Filter
from .kalman_smoother import Smoother
from . import kalman_filter
from . import kalman_smoother
from . import utils
__all__ = []
__all__.extend(kalman_filter.__all__)
__all__.extend(kalman_smoother.__all__)
__all__.append('utils')




import os
import sys
from setuptools import setup, find_packages


if sys.version_info[:2] < (3, 5):
    raise RuntimeError("Python version >= 3.5 required.")

current_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(current_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "linkalman", 
    version = '0.11.5',
    author = 'Danyang Su',
    author_email = 'fnosdy@gmail.com',
    url = 'https://github.com/DanyangSu/linkalman',
    download_url = 'https://github.com/DanyangSu/linkalman/archive/v0.11.2.tar.gz',
    description = 'Flexible Linear Kalman Filter',
    long_description = long_description,
    long_description_content_type='text/markdown',
    keywords = ['kalman', 'time series', 'signal', 'filter'],
    license = '3-clause BSD',
    packages = find_packages(),
    install_requires = [
        'numpy',
        'scipy',
        'pandas',
        'networkx'
        ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
        ]
    )



from collections.abc import Sequence
import numpy as np
import pandas as pd
import networkx as nx
from scipy import linalg
from scipy.linalg import solve_discrete_lyapunov as lyap
from typing import List, Any, Callable, Dict, Tuple
from pandas.api.types import is_numeric_dtype
from numpy.random import multivariate_normal
from copy import deepcopy
import warnings
warnings.simplefilter('default')

__all__ = ['mask_nan', 'inv', 'df_to_list', 'list_to_df', 'create_col', 'get_diag',
        'noise', 'simulated_data', 'gen_PSD', 'ft', 'M_wrap', 'LL_correct', 
        'clean_matrix', 'get_ergodic', 'min_val', 'max_val', 'inf_val', 'pdet',
        'permute', 'revert_permute', 'partition_index', 'get_init_mat', 
        'check_consistence', 'gen_Xt', 'preallocate', 'get_explosive_diffuse',
        'get_nearest_PSD', 'Constant_M']


max_val = 1e6  # Smaller value identifies diffuse better
inf_val = 1e50  # value used to indicate infinity
min_val = 1e-7  # detect 0
    

def mask_nan(is_nan: np.ndarray, mat: np.ndarray, 
        dim: str='both', diag: float=0) -> np.ndarray:
    """
    Takes the list of NaN indices and mask the rows and columns 
    of a matrix with 0 if index has NaN value.

    Example:
    ----------
    mat_masked = mask_nan(np.array([False, True, False]), 
        np.array([[1, 2, 3], 
                [4, 5, 6], 
                [7, 8, 9]]))
    print(mat_masked)  # np.array([[1, 0, 3], [0, 0, 0], [7, 0, 9]])

    Parameters:
    ----------
    mask_nan : indicates if the row and column should be masked
    mat : matrix to be transformed
    dim : whether zero out column or row or both
    diag : replace diagonal NaN with other values

    Returns:
    ----------
    mat_masked : masked mat
    """
    if dim not in ['col', 'row', 'both']:
        raise ValueError('dim must be either "col", "row", or "both"')
    mat_masked = mat.copy()

    # Perform row operation if dim=='row' and 'both' and not a vector
    if dim != 'col' and mat_masked.shape[0] > 1: 
        mat_masked[is_nan] = 0

    # Perform column operation if dim=='col' or 'both' and not a vector
    if dim != 'row' and mat_masked.shape[1] > 1:
        mat_masked[:, is_nan] = 0

    # Replace diagonal values
    if diag != 0:
        for i in range(len(is_nan)):
            if is_nan[i]:
                mat_masked[i][i] = diag

    return mat_masked


def get_diag(arrays: List[np.ndarray]) -> List[np.ndarray]:
    """
    Convert a list of square arrays into a lit of 1D arrays with 
    diagonal values only

    Parameters:
    ----------
    arrays : list of square arrays

    Returns:
    ----------
    diag_arrays : list of N x 1 arrays with diagonal values only
    """
    len_array = len(arrays)
    diag_arrays = preallocate(len_array)
    for i in range(len_array):
        diag_arrays[i] = np.diag(arrays[i]).reshape(-1, 1)
    
    return diag_arrays
    

def inv(h_array: np.ndarray) -> np.ndarray:
    """
    Calculate pinvh of PSD array. Note pinvh performs poorly
    if input matrix is far from being Hermitian, so use pinv2
    instead in this case.

    Parameters:
    ----------
    h_array : input matrix, assume to be Hermitian
    
    Returns:
    ----------
    h_inv : pseudo inverse of h_array. 
    """
    if np.allclose(h_array, h_array.T):
        h_inv = linalg.pinvh(h_array)
    else:
        h_inv = linalg.pinv2(h_array)
    return h_inv


def get_explosive_diffuse(F: np.ndarray) -> List[bool]:
    """
    If contains explosive roots, find strongly
    connected components. Check eigenvalues for
    each component submatrix, set large number to
    the ones with large component and run through 
    Lyapunov solver again to identify all the explosive
    roots.
    
    Parameters:
    ----------
    F : input state transition matrix
    
    Returns:
    ----------
    is_diffuse : indicator of diffuse states
    """
    # Find edges
    nonzeros = np.nonzero(F)
    index_ = list(zip(nonzeros[1], nonzeros[0]))
    
    # Calculate strongly connected components (scc)
    DG = nx.DiGraph()
    DG.add_edges_from(index_)
    scc = list(nx.strongly_connected_components(DG))
    F_ = deepcopy(F)
    
    # Loop through each component for explosive roots
    for comp_ in scc:
        l_comp = list(comp_)
        sub_F = F_[np.ix_(l_comp, l_comp)]
        eig_ = linalg.eigvals(sub_F)
        if np.any(np.abs(eig_) > 1 + min_val):
            F_[l_comp, l_comp] = inf_val
    
    # Get diffuse states 
    is_diffuse = (np.diag(F_) > max_val).tolist()
    return is_diffuse


def gen_Xt(Xt: List[np.ndarray]=None, B: np.ndarray=None, T: int=None) \
        -> List[np.ndarray]:
    """
    Generate a list of zero arrays if Xt is None

    Parameters:
    ----------
    Xt : input Xt
    B : provide dimension information for generating dummy Xt
    T : provide length of the output list Xt

    Returns:
    ----------
    Xt_ : output Xt, if input is None, fill with zeros
    """
    if Xt is None:
        if B is None or T is None:
            raise ValueError('B and T must not be None')
        Xt_ = Constant_M(np.zeros(
                (B.shape[1], 1)), T)
    else:
        Xt_ = Xt
    return Xt_


def df_to_list(df: pd.DataFrame, col_list: List[str]=None) \
        -> List[np.ndarray]:
    """
    Convert pandas dataframe to list of arrays.
    
    Parameters:
    ----------
    df : must be numeric
    col_list : list of columns to be converted

    Returns:
    ----------
    L : len(L) == df.shape[0], L[0].shape[0] == df.shape[1]
    """
    # If col_list is None, return None. None is treated by
    # downstream functions as an indicator.
    if col_list is None:
        return None
    else:

        # Raise exception if df is not a dataframe
        if not isinstance(df, pd.DataFrame):
            raise TypeError('df must be a dataframe')

        # Check data types, must be numeric
        for col in col_list:
            if not is_numeric_dtype(df[col]):
                raise TypeError('Input dataframe must be numeric')

        # Convert df to list row-wise
        L = preallocate(df.shape[0])
        for i in range(df.shape[0]):
            L[i] = np.array([df[col_list].iloc[i, :]]).T
        return L


def list_to_df(L: List[np.ndarray], col: List[str]) -> pd.DataFrame:
    """
    Convert list of arrays to a dataframe. Reverse operation of _df_to_list.

    Parameters:
    ----------
    L : len(L) == df.shape[0], L[0].shape[0] == df.shape[1]
    col : list of column names. len(col) must equal to L[0].shape[0]

    Returns:
    ----------
    df: output dataframe
    """
    if not isinstance(col, list):
        raise TypeError('col must be a list of strings')
    num_col = len(col)
    
    for i in L:
        # Check convertibility
        if i.ndim > 1:
            if i.shape[0] > 1 and i.shape[1] > 1:
                raise ValueError('Input must be one dimensional.')
            else:
                i = i.flatten()

        if num_col != i.shape[0]:
            raise ValueError('Input arrays have wrong size.')

    df_val = np.concatenate([i.T for i in L])
    df = pd.DataFrame(data=df_val, columns=col)
    return df


def create_col(col: List[str], suffix: str='_pred') -> List[str]:
    """
    Create column names for predictions. Default suffix is '_pred'

    Parameters:
    ----------
    col : column list of a dataframe
    suffix : string to be appended to each column name in col

    Returns:
    ----------
    col_new : modified column names
    """
    col_new = []
    for i in col:
        col_new.append(i + suffix)
    return col_new


def preallocate(dim1, dim2=None):
    """
    Preallocate a list by creating either [None] * dim1
    or [[None] * dim2] * dim1. I use for loop to break reference

    Parameters:
    ----------
    dim1 : length of list
    dim2 : if not None, each element is also a list of None

    Returns:
    ----------
    allocated_list : pre-allocated placeholders
    """
    allocated_list = None
    if dim1 > 0:
        if dim2 is None:
            allocated_list = [None for _ in range(dim1)]
        else:
            allocated_list = [[None for _1 in range(dim2)] for _2 in range(dim1)]
    return allocated_list


def noise(y_dim: int, Sigma: np.ndarray) -> np.ndarray:
    """
    Generate n-by-1 Gaussian noise 

    Parameters: 
    ----------
    y_dim : dimension of yt
    Sigma : cov matrix of Gaussian noise

    Returns:
    ----------
    epsilon : noise of the system
    """
    epsilon = multivariate_normal(np.zeros(y_dim), Sigma).reshape(-1, 1)
    return epsilon


def check_consistence(Mt: Dict, y_t: np.ndarray, x_t: np.ndarray, 
        init_state: Dict=None) -> None:
    """
    Check consistence of matrix dimensions. Ensure
    all matrix operations are properly done. The
    assumption is the shape of a matrix remains 
    constant over time. Also assume Mt contains all 
    the required keys. 

    Parameters:
    ----------
    Mt : Dict of system matrices
    y_t : measurement vector
    x_t : regressor vector
    init_state : user-specified initial state values
    """
    # Check whether Mt contains all elements
    keys = set(['Ft', 'Bt', 'Qt', 'Ht', 'Dt', 'Rt', 
            'xi_1_0', 'P_1_0'])
    M_keys = set(Mt.keys())

    if not keys.issubset(M_keys):
        diff_keys = keys.difference(M_keys)
        raise ValueError('Mt does not contain all required keys. ' + \
                'The missing keys are {}'.format(list(diff_keys)))

    # Collect dimensional information
    dim = {}
    dim.update({'Ft': Mt['Ft'][0].shape})
    dim.update({'Bt': Mt['Bt'][0].shape})
    dim.update({'Ht': Mt['Ht'][0].shape})
    dim.update({'Dt': Mt['Dt'][0].shape}) 
    dim.update({'Qt': Mt['Qt'][0].shape}) 
    dim.update({'Rt': Mt['Rt'][0].shape})
    dim.update({'xi_t': Mt['xi_1_0'].shape})
    dim.update({'y_t': y_t.shape}) 
    dim.update({'x_t': x_t.shape})
    
    # Check whether dimension is 2-D
    for m_name in dim.keys():
        if len(dim[m_name]) != 2:
            raise ValueError('{} has the wrong dimensions'.format(m_name))

    # Check Ft and xi_t
    if (dim['Ft'][1] != dim['Ft'][0]) or (dim['Ft'][1] != dim['xi_t'][0]):
        raise ValueError('Ft and xi_t do not match in dimensions')

    # Check Ht and xi_t
    if dim['Ht'][1] != dim['xi_t'][0]:
        raise ValueError('Ht and xi_t do not match in dimensions')

    # Check Ht and y_t
    if dim['Ht'][0] != dim['y_t'][0]:
        raise ValueError('Ht and y_t do not match in dimensions')

    # Check Bt and xi_t
    if dim['Bt'][0] != dim['xi_t'][0]:
        raise ValueError('Bt and xi_t do not match in dimensions')

    # Check Bt and x_t
    if dim['Bt'][1] != dim['x_t'][0]:
        raise ValueError('Bt and x_t do not match in dimensions')

    # Check Dt and y_t
    if dim['Dt'][0] != dim['y_t'][0]:
        raise ValueError('Dt and y_t do not match in dimensions')

    # Check Dt and x_t
    if dim['Dt'][1] != dim['x_t'][0]:
        raise ValueError('Dt and x_t do not match in dimensions')

    # Check Qt and xi_t
    if (dim['Qt'][1] != dim['Qt'][0]) or (dim['Qt'][1] != dim['xi_t'][0]):
        raise ValueError('Qt and xi_t do not match in dimensions')

    # Check Rt and y_t
    if (dim['Rt'][1] != dim['Rt'][0]) or (dim['Rt'][1] != dim['y_t'][0]):
        raise ValueError('Rt and y_t do not match in dimensions')

    # Check if y_t is a vector
    if dim['y_t'][1] != 1:
        raise ValueError('y_t must be a vector')

    # Check if xi_t is a vector
    if dim['xi_t'][1] != 1:
        raise ValueError('xi_t must be a vector')

    # Check if x_t is a vector
    if dim['x_t'][1] != 1:
        raise ValueError('x_t must be a vector')

    # If init_state is provided, check consistency
    if init_state is not None:
        check_name = set(['xi_t','P_star_t', 'P_inf_t']).intersection(
                init_state.keys())
        for name in check_name:

            # Check number of dimensions
            if len(init_state[name].shape) != 2:
                raise ValueError('User-specified {} '.format(name) + \
                        'does not have 2 dimensions')
            
            # Check if match sizes
            if name == 'xi_t':
                dim_check = dim['xi_t']
            else:
                dim_check = dim['Qt']
            if init_state[name].shape != dim_check:
                raise ValueError('User-specified {} has'.format(name) + \
                        ' wrong dimensions')

        if init_state.get('q', 0) < 0:
            raise ValueError('User-specified q must be non-negative')


def simulated_data(Ft: Callable, theta: np.ndarray, 
        Xt: pd.DataFrame=None, T: int=None, init_state: Dict=None, 
        **kwargs) -> Tuple[pd.DataFrame, List[str], List[str]]:
    """
    Generate simulated data from a given BSTS system. Xt and T
    cannot be both set to None. If P_1_0 and xi_1_0  are 
    not provided, calculate ergodic values from system matrices

    Parameters: 
    ----------
    Ft : ft(theta, T) that returns [Mt]_{1,...,T}
    theta : argument of ft
    Xt : input Xt. Optional and can be set to None
    T : length of the time series
    init_state : user-specified initial state values
    kwargs : kwargs for Ft

    Returns:
    ----------
    df : output dataframe that contains Xi_t, Y_t and X_t
    y_col : column names of y_t
    xi_col : column names of xi_t
    """
    # Get dimensional information
    M_ = Ft(theta, 1, **kwargs)
    xi_dim = M_['Ft'][0].shape[0]
    y_dim = M_['Ht'][0].shape[0]
    x_dim = M_['Dt'][0].shape[1]
    
    # Set Xt to Constant_M(np.zeros((1, 1))) if set as None
    if Xt is None:
        if T is None:
            raise ValueError('When Xt = None, T must be assigned')
        X_t = Constant_M(np.zeros((x_dim, 1)), T)
    else:
        T = Xt.shape[0]
        X_t = df_to_list(Xt, list(Xt.columns))
    
    # Check consistence
    x_dim = X_t[0].shape[0]
    y_sample = np.ones([y_dim, 1])
    x_sample = np.ones([x_dim, 1])
    check_consistence(M_, y_sample, x_sample, init_state=init_state)
    
    Mt = Ft(theta, T, **kwargs)
    
    # Generate initial values
    xi_1_0 = deepcopy(Mt['xi_1_0'])
    P_1_0 = deepcopy(Mt['P_1_0'])

    # Override if init_state is provided
    if init_state is not None:
        xi_1_0 = init_state.get('xi_t', xi_1_0)
        P_1_0 = init_state.get('P_star_t', P_1_0)

    P_1_0[np.isnan(P_1_0)] = 1  # give an arbitrary value to diffuse priors
    Xi_t = preallocate(T)
    Xi_t[0] = xi_1_0 + noise(xi_dim, P_1_0)

    # Iterate through time steps
    Y_t = [None] * T
    for t in range(T):
        # Generate y_t
        Y_t[t] = Mt['Ht'][t].dot(Xi_t[t]) + Mt['Dt'][t].dot(X_t[t]) + \
                noise(y_dim, Mt['Rt'][t])

        # Generate xi_t1
        if t < T - 1:
            Xi_t[t+1] = Mt['Ft'][t].dot(Xi_t[t]) + Mt['Bt'][t].dot(X_t[t]) + \
                    noise(xi_dim, Mt['Qt'][t])

    # Generate df
    y_col = ['y_{}'.format(i) for i in range(y_dim)]
    xi_col = ['xi_{}'.format(i) for i in range(xi_dim)]
    df_Y = list_to_df(Y_t, y_col)
    df_Xi = list_to_df(Xi_t, xi_col)
    df = pd.concat([df_Xi, df_Y, Xt], axis=1)
    return df, y_col, xi_col


def gen_PSD(theta: np.ndarray, dim: int) -> np.ndarray:
    """
    Generate covariance matrix from theta. Requirement:
    len(theta) = (dim**2 + dim) / 2

    Parameters:
    ----------
    theta : parameters used in generating lower triangle matrix
        Diagonal values are always non-negative
    dim : dimension of the matrix. 

    Returns:
    PSD : PSD matrix
    """
    # Raise exception if theta has wrong size
    if len(theta) != (dim ** 2 + dim)/2:
        raise ValueError('theta has wrong length')

    L = np.zeros([dim, dim])
    idx = np.tril_indices(dim, k=0)
    L[idx] = theta
    
    #enforce non-negative diagonal values
    for i in range(dim):
        L[i][i] = np.exp(L[i][i])
    PSD = L.dot(L.T)
    return PSD


def get_ergodic(F: np.ndarray, Q: np.ndarray, B: np.ndarray=None,
        x_0: np.ndarray=None, force_diffuse: List[bool]=None, 
        is_warning: bool=True) -> List[np.ndarray]:
    """
    Calculate initial state covariance matrix, and identify 
    diffuse state. It effectively solves a Lyapuov equation

    Parameters:
    ----------
    F : state transition matrix
    Q : initial error covariance matrix
    B : regression matrix
    x_0 : initial x, used for calculating ergodic mean
    force_diffuse : List of booleans of user-determined diffuse state
    is_warning : whether to show warning message

    Returns:
    ----------
    P_0 : the initial state covariance matrix, np.inf for diffuse state
    xi_0 : the initial state mean, 0 for diffuse state
    """
    Q_ = deepcopy(Q)
    dim = Q.shape[0]
    
    # Is is_diffuse is not supplied, create the list
    if force_diffuse is None:
        is_diffuse = [False for _ in range(dim)]
    else: 
        is_diffuse = deepcopy(force_diffuse)
        if len(is_diffuse) != dim:
            raise ValueError('is_diffuse has wrong size')

    # Check F and Q
    if F.shape[0] != F.shape[1]:
        raise TypeError('F must be a square matrix')
    if Q.shape[0] != Q.shape[1]:
        raise TypeError('Q must be a square matrix')
    if F.shape[0] != Q.shape[0]:
        raise TypeError('Q and F must be of same size')
   
    # If explosive roots, use fully diffuse initialization
    # and issue a warning
    eig = linalg.eigvals(F)
    if np.any(np.abs(eig) > 1):
        if is_warning:
            warnings.warn('Ft contains explosive roots. Assumptions ' + \
                    'of marginal LL correction may be violated, and ' + \
                    'results may be biased or inconsistent. Please provide ' + \
                    'user-defined xi_1_0 and P_1_0.', RuntimeWarning)
        is_diffuse_explosive = get_explosive_diffuse(F)
        is_diffuse = [a or b for a, b in zip(is_diffuse, 
            is_diffuse_explosive)]

    # Modify Q_ to reflect diffuse states
    Q_ = mask_nan(is_diffuse, Q_, diag=inf_val)
        
    # Calculate raw P_0
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        P_0 = lyap(F, Q_, 'bilinear')

    # Clean up P_0
    for i in range(dim):
        if np.abs(P_0[i][i]) > max_val:
            is_diffuse[i] = True
    P_0 = mask_nan(is_diffuse, P_0, diag=0)
    
    # Enforce PSD
    P_0_PSD = get_nearest_PSD(P_0)

    # Add nan to diffuse diagonal values
    P_0_PSD += np.diag(np.array([np.nan if i else 0 for i in is_diffuse]))

    # Compute ergodic mean
    if B is None or x_0 is None:
        Bx = np.zeros([dim, 1])
    else:
        if B.shape[0] != dim:
            raise ValueError('B has the wrong dimension')
        Bx = B.dot(x_0)
    Bx[is_diffuse] = 0
    F_star = deepcopy(F)
    F_star[is_diffuse] = 0
    xi_0 = inv(np.eye(dim) - F_star).dot(Bx)

    return P_0_PSD, xi_0


def get_init_mat(P_1_0: np.ndarray) \
        -> Tuple[int, np.ndarray, np.ndarray, np.ndarray]:
    """
    Get information for diffuse initialization

    Parameters:
    ----------
    P_1_0 : initial state covariance. Diagonal value np.nan if diffuse

    Returns:
    ----------
    number_diffuse : number of diffuse state
    A : selection matrix for diffuse states, equal to P_inf
    Pi : selection matrix for stationary states
    P_star : non-diffuse part of P_1_0
    """
    is_diffuse = np.isnan(P_1_0.diagonal())
    number_diffuse = np.count_nonzero(is_diffuse)
    A = np.diag(is_diffuse.astype(float))
    Pi = np.diag((~is_diffuse).astype(float))

    P_clean = deepcopy(P_1_0)
    P_clean[np.isnan(P_clean)] = 0
    P_star = Pi.dot(P_clean).dot(Pi.T)
    
    return number_diffuse, A, Pi, P_star


def get_nearest_PSD(mat: np.ndarray) -> np.ndarray:
    """
    Get the nearest PSD matrix of an input matrix

    Parameters:
    ----------
    mat : input matrix to be processed

    Returns:
    ----------
    PSD_mat : the nearest PSD matrix
    """
    _, S, VT = linalg.svd(mat)
    Sigma = VT.T.dot(np.diag(S)).dot(VT)
    PSD_mat = (Sigma + Sigma.T + mat + mat.T) / 4
    return PSD_mat


def clean_matrix(mat: np.ndarray) -> np.ndarray:
    """
    Enforce 0 and inf on matrix values.
    Convert type to flot if int

    Parameters:
    ----------
    mat : input matrix to be cleaned

    Returns:
    ----------
    cleaned_mat : processed matrix
    """
    cleaned_mat = deepcopy(mat).astype(float, copy=True)
    cleaned_mat[np.abs(cleaned_mat) < min_val] = 0
    cleaned_mat[np.abs(cleaned_mat) > max_val] = inf_val

    return cleaned_mat


def permute(matrix: np.ndarray, index: np.ndarray, 
        axis: str='row') -> np.ndarray:
    """
    Permute a square matrix, may perform row permutation,
    column permutation, or both.
    
    For example:
    a = np.array([[1, 2, 3],
                  [2, 5, 6],
                  [3, 6, 9]])
    b = np.array([2, 0, 1])

    permute(a, b, axis='both') returns:
    np.array([[9, 3, 6],
              [3, 1, 2],
              [6, 2, 5]])

    Parameters:
    ----------
    matrix : input matrix to be permuted
    index : index order for permutation
    axis : axis along which to do permutation

    Returns:
    ----------
    perm_matrix : permuted matrix
    """
    if axis == 'row':
        perm_matrix = matrix[index,:]
    elif axis == 'col':
        perm_matrix = matrix[:, index]
    elif axis == 'both':
        perm_matrix = (matrix[index, :])[:, index]
    else:
        raise ValueError('axis must be "row", "col", or "both".')
    return perm_matrix


def revert_permute(index: np.ndarray) -> np.ndarray:
    """
    Revert matrix indexing to the original order. 

    For example:
    a = np.array([2, 3, 1, 0]) means the original matrix is 
    reordered as [3rd, 4th, 2nd, 1st]. 
    revert_permute(a) returns np.array([3, 2, 0, 1]), which 
    can be used to restore the matrix to its original order.

    Parameters:
    ----------
    index : current indexing

    Returns:
    ----------
    revert_index : index of the original order
    """
    revert_index = index.argsort()
    return revert_index


def partition_index(is_missing: np.ndarray) -> np.ndarray:
    """
    Reshuffle the index with index of observed measurement 
    first. 

    For example:
    is_missing = [True, False, True, False]
    partition_index(is_missing) returns:
    np.array([1, 3, 0, 2])

    Parameters:
    ----------
    is_missing : list of dummies whether y_{t:i} is missing

    Returns:
    ----------
    partitioned_index : partitioned index
    """
    partitioned_index = is_missing.argsort()
    return partitioned_index


def ft(theta: np.ndarray, f: Callable, T: int, x_0: np.ndarray=None, 
        xi_1_0: np.ndarray=None, P_1_0: np.ndarray=None, 
        force_diffuse: List[bool]=None, is_warning: bool=True) -> Dict:
    """
    Duplicate arrays in M = f(theta) and generate list of Mt
    Output of f(theta) must contain all the required keys.

    Parameters:
    ----------
    theta : input of f(theta). Underlying parameters to be optimized
    f : obtained from get_f. Mapping theta to M
    T : length of Mt. "Duplicate" M for T times
    x_0 : establish initial state mean
    xi_1_0 : specify initial state mean. override calculated mean
    P_1_0 : initial state cov
    force_diffuse : use-defined diffuse state
    is_warning : whether to display the warning about explosive roots

    Returns:
    ----------
    Mt : system matrices for a BSTS. Should contain
        all the required keywords. 
    """ 
    M = f(theta)

    # If a value is close to 0 or inf, set them to 0 or inf
    # If an array is 1D, convert it to 2D
    for key in M.keys():
        if M[key].ndim < 2:
            M[key] = M[key].reshape(1, -1)
        M[key] = clean_matrix(M[key])
    
    # Check validity of M
    required_keys = set(['F', 'H', 'Q', 'R'])
    M_keys = set(M.keys())
    if len(required_keys - M_keys) > 0:
        raise ValueError('f does not have required outputs: {}'.
            format(required_keys - M_keys))

    # Check dimensions of M
    for key in M_keys:
        if len(M[key].shape) < 2:
            raise TypeError('System matrices must be 2D')

    # Check PSD of R and Q
    if np.array_equal(M['Q'], M['Q'].T):
        eig_Q = linalg.eigvals(M['Q'])
        if not np.all(eig_Q >= 0):
            raise ValueError('Q is not semi-PSD')
    else: 
        raise ValueError('Q is not symmetric')
    
    if np.array_equal(M['R'], M['R'].T):
        eig_R = linalg.eigvals(M['R'])
        if not np.all(eig_R >= 0):
            raise ValueError('R is not semi-PSD')
    else: 
        raise ValueError('R is not symmetric')
    
    # Generate ft for required keys
    Ft = Constant_M(M['F'], T)
    Ht = Constant_M(M['H'], T)
    Qt = Constant_M(M['Q'], T)
    Rt = Constant_M(M['R'], T)

    # Set Bt if Bt is not Given
    if 'B' not in M_keys:
        dim_xi = M['F'].shape[0]
        if 'D' not in M_keys:
            M.update({'B': np.zeros((dim_xi, 1))})
        else:
            dim_x = M['D'].shape[1]
            M.update({'B': np.zeros((dim_xi, dim_x))})

    if 'D' not in M_keys:
        dim_x = M['B'].shape[1]  # B is already defined
        dim_y = M['H'].shape[0]
        M.update({'D': np.zeros((dim_y, dim_x))})

    # Get Bt and Dt for ft
    Bt = Constant_M(M['B'], T)
    Dt = Constant_M(M['D'], T)

    # Initialization
    if P_1_0 is None or xi_1_0 is None: 
        P_1_0, xi_1_0 = get_ergodic(M['F'], M['Q'], M['B'], 
                x_0=x_0, force_diffuse=force_diffuse, 
                is_warning=is_warning) 
    Mt = {'Ft': Ft, 
            'Bt': Bt, 
            'Ht': Ht, 
            'Dt': Dt, 
            'Qt': Qt, 
            'Rt': Rt,
            'xi_1_0': xi_1_0,
            'P_1_0': P_1_0}
    return Mt


def pdet(array: np.ndarray) -> float:
    """
    Calculate pseudo-determinant. If zero matrix, determinant is 1
    Because we are using log, determinant of 1 is good.

    Parameters:
    ----------
    array : input array
    
    Returns:
    ----------
    array_pdet : pseudo-determinant
    """
    eig, _ = np.linalg.eigh(array)

    # If all eigenvalues are close to 0, np.product(np.array([])) returns 1
    array_pdet = np.product(eig[np.abs(eig) > min_val])
    return array_pdet


def LL_correct(Ht: List[np.ndarray], Ft: List[np.ndarray], 
        n_t: List[int], A: np.ndarray, index: List[List[int]]=None) \
        -> np.ndarray:
    """
    Calculate Correction term for the marginal likelihood

    Parameters:
    ----------
    Ht : list of measurement specification matrices
    Ft : list of state transition matrices
    n_t : only the first n_t[t] rows of Ht are used
    A : selection matrix for P_inf_1_0
    index : if not None, sort Ht first

    Returns:
    MLL_correct : correction term for the marginal likelihood
    """
    psi = deepcopy(A)
    MLL_correct = np.zeros(Ft[0].shape)
    if index is None:
        for t in range(len(Ht)):
            if n_t[t] > 0:
                Zt = Ht[t][0:n_t[t]].dot(psi)
                MLL_correct += (Zt.T).dot(Zt)
            psi = Ft[t].dot(psi)
    else:
        for t in range(len(Ht)):
            if n_t[t] > 0:
                Zt = Ht[t][index[t]][0:n_t[t]].dot(psi)
                MLL_correct += (Zt.T).dot(Zt)
            psi = Ft[t].dot(psi)

    return MLL_correct
    

class M_wrap(Sequence):
    """
    Wraper of array lists. Improve efficiency by skipping 
    repeated calculation when m_list contains same arrays. 
        raise TypeError('Q must be a square matrix')
    """

    def __init__(self, m_list: List[np.ndarray]) -> None:
        """
        Create placeholder for calculated matrix. 

        Parameters:
        ----------
        m_list : list of input arrays. Should be mostly constant
        """
        self.m = None
        self.m_pinvh = None
        self.L = None
        self.D = None
        self.L_I = None
        self.m_pdet = None
        self.m_list = m_list
        

    def __getitem__(self, index: int) -> np.ndarray:
        """
        Returns indexed array of the wrapped list

        Parameters:
        ----------
        index : index of the wrapped list

        Returns:
        ----------
        self.m_list[index] : indexed array of the wrapped list
        """
        return self.m_list[index]


    def __setitem__(self, index: int, val: np.ndarray) -> None:
        """
        Set values of the wrapped list

        Parameters:
        ----------
        index : index of the wrapped list
        val : input array
        """
        self.m_list[index] = val 


    def __len__(self) -> int:
        """
        Required for a Sequence Object

        Returns:
        ----------
        len(self.m_list) : length of the wrapped list
        """
        return len(self.m_list)


    def _equal_M(self, index: int) ->bool:
        """
        Return true if self.m_list[index] == self.m. 
        If false, set self.m = self.m_list[index]

        Parameters: 
        ----------
        index : index of the wrapped list

        Returns:
        ----------
        Boolean that indicates whether we need to perform the operation
        """
        if np.array_equal(self.m, self.m_list[index]):
            return True
        else:
            self.m = self.m_list[index]
            return False
    

    def pinvh(self, index: int) -> np.ndarray:
        """
        Return pseudo-inverse of self.m_list[index]

        Parameters:
        ----------
        index : index of the wrapped list

        Returns:
        self.m_pinvh : pseudo-inverse
        """
        if (not self._equal_M(index)) or self.m_pinvh is None:
            self.m_pinvh = inv(self.m)
        return self.m_pinvh
    

    def ldl(self, index: int) -> List[np.ndarray]:
        """
        Calculate L and D from LDL decomposition, and inverse of L

        Parameters:
        ----------
        index : index of the wrapped list

        Returns:
        self.L : L  of LDL
        self.D : D of LDL
        self.L_I : inverse of L
        """
        if (not self._equal_M(index)) or self.L is None:
            self.L, self.D, _ = linalg.ldl(self.m)
            self.L_I, _ = linalg.lapack.dtrtri(self.L, lower=True)
        return self.L, self.D, self.L_I


    def pdet(self, index: int) -> float:
        """
        Calculate pseudo-determinant.

        Parameters:
        ----------
        index : index of the wrapped list
        
        Returns:
        ----------
        self.m_pdet : pseudo-determinant
        """
        if (not self._equal_M(index)) or self.m_pdet is None:
            self.m_pdet = pdet(self.m)
        return self.m_pdet


class Constant_M(Sequence):
    """
    If the sequence of system matrix is mostly constant over time 
    (with the exception of occasional deviation), using Constant_M
    saves memory space. It mimics the behavior of a regular list but 
    use one single baseline M and stores any deviation.

    Example:
    ----------
    M = np.array([[5, 3],[3, 4]])  # baseline matrix
    T = 100  # intended length of the list
    
    # Mt has similar behavior as [copy.deepcopy(M) for _ in range(T)]
    Mt = Constant_M(M, T)  
    
    # Modify Mt[i] for i doesn't affect Mt[j] for j!=i
    Mt[2] = np.array([[5, 2], [2, 5]])
    print(Mt[2])  # np.array([[5, 2], [2, 5]])
    print(Mt[1])  # np.array([[5, 3], [3, 4]])
        raise TypeError('Q must be a square matrix')
    """

    def __init__(self, M: np.ndarray, length: int) -> None:
        """
        Generate a list of M.

        Parameters:
        ----------
        M : input system matrix 
        length : length of the list
        """
        # Raise if M is not np.array type
        if not isinstance(M, np.ndarray):
            raise TypeError('M must be a numpy array')

        self.M = deepcopy(M)
        self._M = deepcopy(M)  # benchmark M
        self.index = None
        self.Mt = {}
        self.length = length


    @property
    def is_M_modified(self) -> bool:
        """
        Determine whether self.M is modified or not

        Returns:
        ----------
        Boolean that returns True if self.M is different from self._M
        """
        return not np.array_equal(self.M, self._M)


    def finger_print(self) -> None:
        """
        Update self.Mt and restore self.M if self.M is modified. This 
        happens when elements of a numpy array is modified and 
        __setitem__ is not invoked. The modified value will be kept 
        in self.M. This function will record the change before restoring
        self.M to its original state.
        """
        self.Mt.update({self.index: deepcopy(self.M)})
        self.M = deepcopy(self._M)


    def __setitem__(self, index: int, val: np.ndarray) -> None:
        """
        If val differ from self.M, store val and index
        in self.Mt. 

        Parameters:
        ----------
        index : index number of the list between 0 and self.T
        val : value to replace M at index. 
        """
        # Check if already partially updated. If it is, update self.Mt
        if self.index == index and self.is_M_modified:
            self.finger_print()

        # Only update if val differs from current value or self._M
        if not np.array_equal(self.Mt.get(index, self._M), val):
            self.Mt.update({index: deepcopy(val)}) 


    def __getitem__(self, index: int) -> np.ndarray:
        """
        Search through self.Mt dictionary, return 
        self.Mt[index] if self.Mt[index] is set, 
        else returns default self.M. In addition,
        if a modification on self.M is detected, 
        update self.Mt and restore self.M

        Parameters:
        ----------
        index : index number of the list

        Returns:
        ----------
        Mt_index : Constant_M[index]
        """
        # Restore self.M and update self.Mt
        if self.is_M_modified:
            self.finger_print()

        Mt_index = self.Mt.get(index, self.M)
        self.index = index
        return Mt_index


    def __len__(self) -> int:
        """
        Set length of the list, required for an object inherited 
        from Sequence.

        Returns:
        ----------
        self.length : length of the list
        """
        return self.length

